import numpy as np
from sklearn.base import BaseEstimator, RegressorMixin, clone, is_regressor, is_classifier
from sklearn.utils.validation import check_is_fitted, check_X_y, check_array
from sklearn.exceptions import NotFittedError
from sklearn.model_selection import train_test_split

class QuantileStackRegressor(BaseEstimator, RegressorMixin):
    """
    A meta regressor for doing model stacking using underlying 
    quantile propensity models. The model will first learn a 
    series of quantile discriminator functions and then stack
    them with out of sample predictions into a final regressor.
 
    Particularly useful for zero-inflated or heavily skewed datasets,

    `QuantileStackRegressor` consists of a series of classifiers and a regressor.

        - The classifier's task is to build a series of propensity models 
          that predict if the target is above a given threshold. 
          These are built in a two fold CV, so that out of sample predictions
          can be added to the x vector for the final regression model
        - The regressor's task is to output the final prediction, aided by the 
          probabilities added by the underlying quantile classifiers.

    At prediction time, the average of the two classifiers is used for all propensity models.

    Credits: This structure of this code is based off the zero inflated regressor from sklego:
             https://github.com/koaning/scikit-lego

    Parameters
    ----------
    classifier : Any, scikit-learn classifier

    regressor : Any, scikit-learn regressor

    Examples
    --------
    >>> import numpy as np
    >>> from sklearn.ensemble import ExtraTreesClassifier, ExtraTreesRegressor
    >>> np.random.seed(0)
    >>> X = np.random.randn(10000, 4)
    >>> y = ((X[:, 0]>0) & (X[:, 1]>0)) * np.abs(X[:, 2] * X[:, 3]**2)
    >>> z = QuantileStackRegressor(
    ... classifier=ExtraTreesClassifier(random_state=0),
    ... regressor=ExtraTreesRegressor(random_state=0)
    ... )
    >>> z.fit(X, y)
    QuantileStackRegressor(classifier=ExtraTreesClassifier(random_state=0),
                          regressor=ExtraTreesRegressor(random_state=0))
    >>> z.predict(X)[:5]
    array([4.91483294, 0.        , 0.        , 0.04941909, 0.        ])
    """

    def __init__(self, classifier, regressor, cuts=[0]) -> None:
        """Initialize."""
        self.classifier = classifier
        self.regressor = regressor
        self.cuts = cuts

    def fit(self, X, y, sample_weight=None):
        """
        Fit the model.

        Parameters
        ----------
        X : np.ndarray of shape (n_samples, n_features)
            The training data.

        y : np.ndarray, 1-dimensional
            The target values.

        sample_weight : Optional[np.array], default=None
            Individual weights for each sample.

        Returns
        -------
        QuantileStackRegressor
            Fitted regressor.

        Raises
        ------
        ValueError
            If `classifier` is not a classifier or `regressor` is not a regressor.
        """
        X, y = check_X_y(X, y)
        self._check_n_features(X, reset=True)

        if not is_classifier(self.classifier):
            raise ValueError(
                f"`classifier` has to be a classifier. Received instance of {type(self.classifier)} instead.")
        if not is_regressor(self.regressor):
            raise ValueError(f"`regressor` has to be a regressor. Received instance of {type(self.regressor)} instead.")

        """
          Now we need to internally split the data and build two sets of the classifiers 
          to prevent target leakage
        """
        X_ = [0] * 2
        y_ = [0] * 2
        X_[0], X_[1], y_[0], y_[1] = train_test_split(X, y, test_size=0.5)

        """
          Build two sets of classifiers for each of the required cuts
        """
        self.classifiers_ = [0] * 2

        for index in [0,1]:
            self.classifiers_[index] = [0] * len(self.cuts)
            for c, cut in enumerate(self.cuts):
                self.classifiers_[index][c] = clone(self.classifier)
                self.classifiers_[index][c].fit(X_[index], y_[index] > cut )

        """
          Apply those classifier to the out of sample data
        """
        Xfinal_ = [0] * 2
        for index in [0,1]:
            Xfinal_[index] = X_[index].copy()
            c_index = 1 - index
            for c, cut in enumerate(self.cuts):
                preds = self.classifiers_[c_index][c].predict_proba( X_[index] )[:,1]
                Xfinal_[index] = np.append(Xfinal_[index], preds.T[:, None], axis=1)

        """
          Join the split data into a final dataset for the regression model
        """
        Xfinale = np.concatenate((Xfinal_[0], Xfinal_[1] ), axis=0) 
        Yfinale = np.concatenate((y_[0], y_[1] ), axis=0)

        self.regressor_ = clone(self.regressor)
        self.regressor_.fit( Xfinale, Yfinale, sample_weight=sample_weight)
        return self


    def predict(self, X):
        """
        Get predictions.

        Parameters
        ----------
        X : np.ndarray, shape (n_samples, n_features)
            Samples to get predictions of.

        Returns
        -------
        y : np.ndarray, shape (n_samples,)
            The predicted values.
        """
        check_is_fitted(self)
        X = check_array(X)
        self._check_n_features(X, reset=False)

        """
          Apply classifiers to generate new colums
        """
        Xfinale = X.copy()

        for c, cut in enumerate(self.cuts):
            temp = np.zeros(len(X))
            for index in [0,1]:
                temp = temp + self.classifiers_[index][c].predict_proba(X)[:,1]
            temp = temp/2
            Xfinale = np.append(Xfinale, temp[:, None], axis=1)

        return self.regressor_.predict(Xfinale)



from sklearn.base import BaseEstimator, RegressorMixin, clone, is_regressor
from sklearn.utils.validation import check_is_fitted, check_X_y, check_array
from sklearn.exceptions import NotFittedError
import pandas as pd
import numpy as np

class BaselineProportionalRegressor(BaseEstimator, RegressorMixin):
    """
    A meta regressor for learning the target value as a proportional difference
       relative to a mean value for a subset of other features.

    Creates and maintains an internal lookup table for the baseline during the
       model fir process.
 
    Parameters
    ----------
    regressor : Any, scikit-learn regressor that will be learned for the adjust target
    """

    def __init__(self, baseline_cols, regressor) -> None:
        """Initialize."""
        self.baseline_cols = baseline_cols
        self.regressor = regressor
        self.baseline_func = 'mean'

    def generate_baseline(self, df):
        self.lookup = df.groupby(self.baseline_cols).agg({'baseline':self.baseline_func}).reset_index()
        self.baseline_default = df['baseline'].agg(self.baseline_func)        

    def get_baseline_predictions(self, df):
        new_df = pd.merge(df, self.lookup, how='left', on=self.baseline_cols)
        new_df['baseline'] = np.where(new_df['baseline'].isnull(), self.baseline_default, new_df['baseline'])
        return new_df['baseline']

    def get_relative_target(self, baseline, y):
        return (y-baseline)/baseline
 
    def invert_relative_target(self, preds, baseline):
        return (preds*baseline)+baseline

    def get_params(self, deep=True):
        return self.regressor.get_params()

    def set_params(self, **parameters):
        for parameter, value in parameters.items():
            if parameter == "baseline_cols":
                self.baseline_cols = value
            else:
                self.regressor.setattr(parameter, value)
        return self

    def fit(self, X, y, sample_weight=None):
        """
        Fit the model.

        Note: this model diverges from the scikit-learn standard in that it needs a 
        pandas dataframe. 

        Parameters
        ----------
        X : pandas.DataFrame of shape (n_samples, n_features)
            The training data.

        y : np.ndarray, 1-dimensional
            The target values.

        sample_weight : Optional[np.array], default=None
            Individual weights for each sample.

        Returns
        -------
        BaselineProportionalRegressor
            Fitted regressor.

        Raises
        ------
        ValueError
            If `regressor` is not a regressor.
        """
        column_names = X.columns
        X, y = check_X_y(X, y)
        self._check_n_features(X, reset=True)
        X = pd.DataFrame(X, columns=column_names)
        
        if not is_regressor(self.regressor):
            raise ValueError(f"`regressor` has to be a regressor. Received instance of {type(self.regressor)} instead.")

        for col in self.baseline_cols:
            if col not in X.columns:
                raise ValueError(f"pandas.DataFrame required with baseline columns: `{col}` NOT FOUND.")

        df = X.copy()
        df['baseline'] = y
        self.generate_baseline(df)
        baseline = self.get_baseline_predictions(X)

        Yfinale = self.get_relative_target(baseline, y)
        self.regressor_ = clone(self.regressor)
        self.regressor_.fit( X, Yfinale, sample_weight=sample_weight)
        return self


    def predict(self, X):
        """
        Get predictions.

        Parameters
        ----------
        X : pd.DataFrame - shape (n_samples, n_features)
            DataFrame of samples to get predictions for.
            Note: DataFrame is required because the baseline uses column names.

        Returns
        -------
        y : np.ndarray, shape (n_samples,)
            The predicted values.
        """
        
        check_is_fitted(self)
        column_names = X.columns
        X = check_array(X)
        self._check_n_features(X, reset=False)
        X = pd.DataFrame(X, columns=column_names)

        for col in self.baseline_cols:
            if col not in X.columns:
                raise ValueError(f"pandas.DataFrame required with baseline columns: `{col}` NOT FOUND.")

        baseline = self.get_baseline_predictions(X)
        preds = self.regressor_.predict(X)

        return self.invert_relative_target(preds, baseline)



from sklearn.base import BaseEstimator, TransformerMixin
import pandas as pd
import numpy as np

class ColumnConcatenator(BaseEstimator, TransformerMixin):
    """
    This is a simple SK learn compatible feature transformer
    which will transform concatenate two columns into one.
    The resulting column will be categorical as the contents
    will be coerced into strings
    """
    def __init__(self, input_column_names):
        self.input_column_names = input_column_names
        if len(self.input_column_names)!=2:
           print("ERROR")
           exit(1)
        self.column_one = self.input_column_names[0]
        self.column_two = self.input_column_names[1]
        self.output_name = self.column_one + "_" + self.column_two
        self.output_names = [self.output_name]

    def fit( self, X, y = None ):
        return self

    def transform(self, X, y=None):
        df = pd.DataFrame(data=X, columns = self.input_column_names)
        def merge_names(row):
            return row[self.column_one]+ "-" + row[self.column_two]
        df[self.output_name] = df.apply(merge_names, axis=1)
        return df.loc[:,self.output_names]



import numpy as np
from sklearn.base import BaseEstimator, RegressorMixin, clone, is_regressor, is_classifier
from sklearn.utils.validation import check_is_fitted, check_X_y, check_array
from sklearn.exceptions import NotFittedError
from sklearn.model_selection import train_test_split

class RegressorStack(BaseEstimator, RegressorMixin):
    """
    A meta regressor for doing model stacking for regression using underlying 
    quantile propensity models and internal regressors.
    Particularly designed for zero-inflated or heavily skewed datasets,

   `RegressorStack` consists of a series of internal regressors
     all of which are fitted in an internal cross validation and scored out-of-sample
     A final regressor is trained over the original features and the output
     of these stacked regression models.

    Parameters
    ----------
    regressor : Any, scikit-learn regressor
    A regressor for predicting the target. 

    Examples
    --------
    >>> import numpy as np
    >>> from sklearn.ensemble import ExtraTreesClassifier, ExtraTreesRegressor
    >>> np.random.seed(0)
    >>> X = np.random.randn(10000, 4)
    >>> y = ((X[:, 0]>0) & (X[:, 1]>0)) * np.abs(X[:, 2] * X[:, 3]**2)
    >>> z = RegressorStack(
    ...   [KNeighborsRegressor(), BayesianRidge()],
    ...   regressor=ExtraTreesRegressor(random_state=0)
    ... )
    >>> z.fit(X, y)
    RegressorStack([KNeighborsRegressor(), BayesianRidge()],
                          regressor=ExtraTreesRegressor(random_state=0))
    >>> z.predict(X)[:5]
    array([4.91483294, 0.        , 0.        , 0.04941909, 0.        ])
    """

    def __init__(self, regressor_list, regressor) -> None:
        """Initialize."""
        self.regressor_list = regressor_list
        self.regressor = regressor

    def fit(self, X, y, sample_weight=None):
        """
        Fit the model.

        Parameters
        ----------
        X : np.ndarray of shape (n_samples, n_features)
            The training data.

        y : np.ndarray, 1-dimensional
            The target values.

        sample_weight : Optional[np.array], default=None
            Individual weights for each sample.

        Returns
        -------
        StackedRegressor
            Fitted regressor.

        Raises
        ------
        ValueError
            If `regressor` is not a regressor.
        """
        X, y = check_X_y(X, y)
        self._check_n_features(X, reset=True)

        if not is_regressor(self.regressor):
            raise ValueError(f"`regressor` has to be a regressor. Received instance of {type(self.regressor)} instead.")

        """
          Now we need to internally split the data and build two sets of
          internal regressors to prevent leakage
        """
        X_ = [0] * 2
        y_ = [0] * 2
        X_[0], X_[1], y_[0], y_[1] = train_test_split(X, y, test_size=0.5)

        """
          Build the internal regressors 
        """
        self.regressors_ = [0] * 2

        for index in [0,1]:
            self.regressors_[index] = [0] * len(self.regressor_list)
            for c, reg in enumerate(self.regressor_list):
                self.regressors_[index][c] = clone(reg)
                self.regressors_[index][c].fit(X_[index], y_[index] )

        """
          Apply those classifier to the out of sample data
        """
        Xfinal_reg_ = [0] * 2
        for index in [0,1]:
            Xfinal_reg_[index] = X_[index].copy()
            c_index = 1 - index
            for c, reg in enumerate(self.regressor_list):
                preds = self.regressors_[c_index][c].predict( X_[index] )
                Xfinal_reg_[index] = np.append(Xfinal_reg_[index], preds.T[:, None], axis=1)

        """
          Join the split data into a final dataset for the regression model
        """
        Xfinale = np.concatenate((Xfinal_reg_[0], Xfinal_reg_[1] ), axis=0) 
        Yfinale = np.concatenate((y_[0], y_[1] ), axis=0)

        self.regressor_ = clone(self.regressor)
        self.regressor_.fit( Xfinale, Yfinale, sample_weight=sample_weight)
        return self


    def predict(self, X):
        """
        Get predictions.

        Parameters
        ----------
        X : np.ndarray, shape (n_samples, n_features)
            Samples to get predictions of.

        Returns
        -------
        y : np.ndarray, shape (n_samples,)
            The predicted values.
        """
        check_is_fitted(self)
        X = check_array(X)
        self._check_n_features(X, reset=False)

        Xfinale = X.copy()

        for c, reg in enumerate(self.regressor_list):
            temp = np.zeros(len(X))
            for index in [0,1]:
                temp = temp + self.regressors_[index][c].predict( X )
            temp = temp/2
            Xfinale = np.append(Xfinale, temp[:, None], axis=1)

        return self.regressor_.predict(Xfinale)



from sklearn.base import BaseEstimator, TransformerMixin
import pandas as pd
import numpy as np

class TwoColumnRatioDiff(BaseEstimator, TransformerMixin):
    """
    This is a simple SK learn compatible feature transformer
    which will transform two numerical columns into their ratio
    and difference
    """
    def __init__(self, input_column_names):
        self.input_column_names = input_column_names
        if len(self.input_column_names)!=2:
           print("ERROR")
           exit(1)
        self.column_one = self.input_column_names[0]
        self.column_two = self.input_column_names[1]
        self.diff_name = self.column_one + "-" + self.column_two
        self.ratio_name = self.column_one + "/" + self.column_two
        self.output_names = [self.ratio_name, self.diff_name]

    def fit( self, X, y = None ):
        return self

    def transform(self, X, y=None):
        df = pd.DataFrame(data=X, columns = self.input_column_names)
        df[self.ratio_name] = np.where(
            df[self.column_two]>0,
            df[self.column_one]/df[self.column_two], 0.0
        )
        df[self.diff_name] = df[self.column_one]-df[self.column_two]
        return df.loc[:,self.output_names]




__all__ = [
    "RegressorStack",
    "QuantileStackRegressor",
    "BaselineProportionalRegressor"
]
from .regressor_stack import RegressorStack
from .quantile_stack_regressor import QuantileStackRegressor
from .baseline_proportional_regressor import BaselineProportionalRegressor



from sklearn.base import BaseEstimator, TransformerMixin

class LookupEncoder(BaseEstimator, TransformerMixin):
    def __init__(self, column_name, lookup_table, default_value):
        self.column_name = column_name
        self.lookup_table = lookup_table
        self.default_value = default_value

    def fit(self, X, y=None):
        return self

    def transform(self, X):
        output_df = X.copy()
        output_df[self.column_name] = output_df[self.column_name].apply(lambda x: self.lookup_table.get(x, self.default_value))
        return output_df


